
class CalAicObj:

    def __init__(self, RSS, tr_S, llf, aa, n):
        self.RSS = RSS
        self.tr_S = tr_S
        self.llf = llf
        self.aa = aa
        self.n = n


class CalMultiObj:

    def __init__(self, betas, reside):
        self.betas = betas
        self.reside = reside
